package edu.uiuc.ncsa.myproxy.oa4mp.server;

import edu.uiuc.ncsa.myproxy.MyProxyServiceFacade;
import edu.uiuc.ncsa.myproxy.ServiceFacadeConfiguration;
import edu.uiuc.ncsa.security.core.configuration.Configurations;
import edu.uiuc.ncsa.security.core.configuration.provider.CfgEvent;
import edu.uiuc.ncsa.security.core.configuration.provider.HierarchicalConfigProvider;
import edu.uiuc.ncsa.security.util.ssl.SSLKeystoreConfiguration;
import org.apache.commons.configuration.tree.ConfigurationNode;

import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.List;

import static edu.uiuc.ncsa.security.core.configuration.Configurations.getFirstAttribute;

/**
 * <p>Created by Jeff Gaynor<br>
 * on 1/19/12 at  4:35 PM
 */
public class MyProxyFacadeProvider extends HierarchicalConfigProvider<MyProxyServiceFacade> implements OA4MPConfigTags {

    @Override
    protected boolean checkEvent(CfgEvent cfgEvent) {
        if (cfgEvent.getConfiguration().getName().equals("myproxy")) {
            setConfig(cfgEvent.getConfiguration());
            return true;
        }
        return false;
    }

    @Override
    public Object componentFound(CfgEvent configurationEvent) {
        if (checkEvent(configurationEvent)) {
            return get();
        }
        return null;
    }

    @Override
    public MyProxyServiceFacade get() {
        ServiceFacadeConfiguration sfc;
        HashMap<String, Integer> loas = new HashMap<String, Integer>();
        String localhostname = null;
        try {
            localhostname = java.net.InetAddress.getLocalHost().getCanonicalHostName();
        } catch (UnknownHostException e) {
            localhostname = "localhost";
        }
        int port = 7512; //start with default port.

        if (getConfig() == null) {
            sfc = new ServiceFacadeConfiguration(localhostname, port, loas);
            //   return new MyProxyServiceFacade(sfc);
        } else {
            try {
                port = getIntAttribute(MYPROXY_PORT);

            } catch (Throwable t) {
                // do nothing. If the port is not given, use the default
            }
            if (getAttribute(MYPROXY_HOST) != null) {
                localhostname = getAttribute(MYPROXY_HOST);
            }

            sfc = new ServiceFacadeConfiguration(localhostname, port, loas);
            List list = getConfig().getChildren(MYPROXY_LOA);
            if (!list.isEmpty()) {
                for (Object obj : list) {
                    ConfigurationNode cn = (ConfigurationNode) obj;
                    loas.put(Configurations.getFirstAttribute(cn, MYPROXY_LOA_NAME),
                            Integer.parseInt(Configurations.getFirstAttribute(cn, MYPROXY_LOA_PORT)));
                }
            }
        }
        // Unless there is something very exotic about your setup, a basic configuration that
        // points to the standard keystore available in java should be more than sufficient.
        SSLKeystoreConfiguration keystoreConfiguration = new SSLKeystoreConfiguration();
        List keystores = null;
        if (getConfig() != null) {
            keystores = getConfig().getChildren(SSL_KEYSTORE);
        }

        if (keystores == null || keystores.isEmpty()) {
            keystoreConfiguration.setUseDefaultJavaKeyStore(true); // default
        } else {
            ConfigurationNode cn2 = (ConfigurationNode) keystores.get(0);
            keystoreConfiguration.setKeystore(getFirstAttribute(cn2, SSL_KEYSTORE_PATH));
            keystoreConfiguration.setKeystorePassword(getFirstAttribute(cn2, SSL_KEYSTORE_PASSWORD));
            keystoreConfiguration.setKeyManagerFactory(getFirstAttribute(cn2, SSL_KEYSTORE_FACTORY));
            keystoreConfiguration.setKeystoreType(getFirstAttribute(cn2, SSL_KEYSTORE_TYPE));
            String x = getFirstAttribute(cn2, SSL_KEYSTORE_USE_JAVA_KEYSTORE);
            if (x == null) {
                keystoreConfiguration.setUseDefaultJavaKeyStore(true); //default
            } else {
                keystoreConfiguration.setUseDefaultJavaKeyStore(Boolean.parseBoolean(x)); //default
            }
        }
        return  new MyProxyServiceFacade(sfc, keystoreConfiguration);
    }

    public MyProxyFacadeProvider() {
    }

    public MyProxyFacadeProvider(ConfigurationNode config) {
        super(config);
    }
}
