package edu.uiuc.ncsa.myproxy.oa4mp.server.servlet;

import edu.uiuc.ncsa.myproxy.oa4mp.server.DSTransaction;
import edu.uiuc.ncsa.security.core.exceptions.GeneralException;
import edu.uiuc.ncsa.security.delegation.server.request.ATRequest;
import edu.uiuc.ncsa.security.delegation.server.request.ATResponse;
import edu.uiuc.ncsa.security.delegation.servlet.TransactionState;
import edu.uiuc.ncsa.security.delegation.token.Verifier;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import static edu.uiuc.ncsa.security.core.util.DateUtils.checkTimestamp;

/**
 * <p>Created by Jeff Gaynor<br>
 * on May 18, 2011 at  11:20:06 AM
 */
public class AccessTokenServlet extends MyProxyDelegationServlet {
    @Override
    protected void doIt(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse) throws Throwable {
        doDelegation(httpServletRequest, httpServletResponse);
    }

    protected TransactionState doDelegation(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse) throws Throwable, ServletException {
        info("5.a. Starting access token exchange");
        Verifier v = getServiceEnvironment().getTokenForge().getVerifier(httpServletRequest);
        debug("5.a. verifier = " + v);
        checkTimestamp(v.getToken());
        DSTransaction transaction = (DSTransaction) getTransactionStore().get(v);
        checkClient(transaction.getClient());
        String cc = "client="+transaction.getClient().getIdentifierString();
        info("5.a. " + cc);
        if (transaction == null) {
            warn("Error: no transaction found for verifier \"" + v + "\"s");
            throw new ServletException("Error: no transaction found for verifier \"" + v + "\"s");
        }
        debug("5.a. grant valid=" + transaction.isAuthGrantValid() + ", at valid=" + transaction.isAccessTokenValid());
        if (!transaction.isAuthGrantValid() || transaction.isAccessTokenValid()) {
            String msg = "Error: the state of the transaction is invalid for auth grant " + transaction.getAuthorizationGrant();
            warn(msg);
            throw new GeneralException(msg);
        }
        preprocess(new TransactionState(httpServletRequest, httpServletResponse, null, transaction));

        ATRequest atRequest = new ATRequest(httpServletRequest, getClient(httpServletRequest));

        info("5.a. got access token " + cc);
        atRequest.setVerifier(v);
        atRequest.setAuthorizationGrant(transaction.getAuthorizationGrant());

        ATResponse atResp = (ATResponse) getATI().process(atRequest);

        debug("5.a. access token = " + atResp.getAccessToken() + " for verifier = " + v);
        transaction.setAuthGrantValid(false);
        transaction.setAccessToken(atResp.getAccessToken());
        transaction.setAccessTokenValid(true);

        try {
            getTransactionStore().save(transaction);
            info("5.a. updated transaction state for " + cc  + ", sending response to client");

        } catch (GeneralException e) {
            throw new ServletException("Error saving transaction", e);
        }
        atResp.write(httpServletResponse);
        info("5.b. done with access token exchange with " + cc);
        TransactionState transactionState =  new TransactionState(httpServletRequest, httpServletResponse, atResp.getParameters(), transaction);
        postprocess(transactionState);
        return transactionState;
    }
}
