package edu.uiuc.ncsa.myproxy.oa4mp.server.servlet;

import edu.uiuc.ncsa.myproxy.MyProxyServiceFacade;
import edu.uiuc.ncsa.myproxy.oa4mp.server.DSException;
import edu.uiuc.ncsa.myproxy.oa4mp.server.DSTransaction;
import edu.uiuc.ncsa.security.core.exceptions.GeneralException;
import edu.uiuc.ncsa.security.core.util.DateUtils;
import edu.uiuc.ncsa.security.delegation.servlet.TransactionState;
import edu.uiuc.ncsa.security.delegation.token.AuthorizationGrant;
import edu.uiuc.ncsa.security.delegation.token.MyX509Certificates;
import edu.uiuc.ncsa.security.servlet.JSPUtil;
import edu.uiuc.ncsa.security.servlet.Presentable;
import edu.uiuc.ncsa.security.servlet.PresentableState;
import edu.uiuc.ncsa.security.util.pkcs.CertUtil;
import org.bouncycastle.jce.PKCS10CertificationRequest;
import org.globus.gsi.X509ExtensionSet;
import org.globus.gsi.bc.BouncyCastleCertProcessingFactory;
import org.globus.gsi.proxy.ext.ProxyCertInfo;
import org.globus.gsi.proxy.ext.ProxyCertInfoExtension;
import org.globus.gsi.proxy.ext.ProxyPolicy;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.KeyPair;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.util.LinkedList;

import static org.apache.commons.lang.StringEscapeUtils.escapeHtml;
import static org.globus.gsi.GSIConstants.CertificateType.GSI_4_LIMITED_PROXY;

/**
 * This servlet handles the authorization logic. It creates/access the appropriate state then forwards this
 * on to a JSP page for presentation.
 * <p>Created by Jeff Gaynor<br>
 * on 9/25/11 at  6:40 PM
 */
public class AuthorizationServlet extends MyProxyDelegationServlet implements Presentable {

    public static final String AUTHORIZATION_ACTION_KEY = "action";
    public static final String AUTHORIZATION_USER_NAME_KEY = "AuthUserName";
    public static final String AUTHORIZATION_USER_NAME_VALUE = "userName"; // only used for setting the username, if it comes in a header.
    public static final String AUTHORIZATION_PASSWORD_KEY = "AuthPassword";
    public static final String AUTHORIZATION_ACTION_OK_VALUE = "ok";
    public static final int AUTHORIZATION_ACTION_OK = 1;
    public static final int AUTHORIZATION_ACTION_START = 0;
    public static final String RETRY_MESSAGE = "retryMessage";

    /**
     * State object after authorization has worked.
     */
    protected class AuthorizedState extends PresentationState {
        public AuthorizedState(int state, HttpServletRequest request, HttpServletResponse response, DSTransaction transaction) {
            super(state, request, response);
            this.transaction = transaction;
        }


        public DSTransaction getTransaction() {
            return transaction;
        }

        DSTransaction transaction;
    }

    public void prepare(PresentableState state) throws Throwable {
        AuthorizedState aState = (AuthorizedState) state;
        switch (aState.getState()) {
            case AUTHORIZATION_ACTION_OK:
                // nothing to do, really
                return;
            case AUTHORIZATION_ACTION_START:
                info("3.a. Starting authorization for grant =" + aState.getTransaction().getIdentifierString());
                //Mess of information for the form
                HttpServletRequest request = state.getRequest();

                request.setAttribute(AUTHORIZATION_USER_NAME_KEY, AUTHORIZATION_USER_NAME_KEY);
                request.setAttribute(AUTHORIZATION_PASSWORD_KEY, AUTHORIZATION_PASSWORD_KEY);
                request.setAttribute(AUTHORIZATION_ACTION_KEY, AUTHORIZATION_ACTION_KEY);
                request.setAttribute("actionOk", AUTHORIZATION_ACTION_OK_VALUE);
                request.setAttribute("authorizationGrant", aState.getTransaction().getIdentifierString());
                request.setAttribute("tokenKey", TOKEN_KEY);
                // HTML escape it to guard against HTML injection attacks. Addresses issue OAUTH-87.
                // If you aren't sure whether a form is secure against HTML injection attacks, paste the following into it:
                // ><script>alert('CSS Vulnerable')</script><b a=a     a></a><script>alert('CSS Vulnerable')</script>     \'><script>alert%28\'CSS Vulnerable\'%29</script><
                // and get the form to re-display. If it is vulnerable, a popup saying so will appear.

                request.setAttribute("clientHome", escapeHtml(aState.getTransaction().getClient().getHomeUri()));
                request.setAttribute("clientName", escapeHtml(aState.getTransaction().getClient().getName()));
                //request.setAttribute("clientName", aState.getTransaction().getClient().getName());
                request.setAttribute("actionToTake", request.getContextPath() + "/authorize");
                return;
        }
    }

    public static String INITIAL_PAGE = "/authorize-init.jsp";
    public static String OK_PAGE = "/authorize-ok.jsp";
    public static String ERROR_PAGE = "/authorize-error.jsp";

    public void present(PresentableState state) throws Throwable {
        AuthorizedState aState = (AuthorizedState) state;
        postprocess(new TransactionState(state.getRequest(), aState.getResponse(), null, aState.getTransaction()));

        switch (aState.getState()) {
            case AUTHORIZATION_ACTION_START:
                String initPage = INITIAL_PAGE;
                info("*** STARTING present");
                if (getServiceEnvironment().getAuthorizationServletConfig().isUseHeader()) {
                    info("*** PRESENT: Use headers enabled.");
                    String x = null;
                    if (getServiceEnvironment().getAuthorizationServletConfig().getHeaderFieldName().equals("REMOTE_USER")) {
                        // slightly more surefire way to get this.
                        x = aState.getRequest().getRemoteUser();
                        info("*** got user name from request = " + x);
                    } else {
                        x = aState.getRequest().getHeader(getServiceEnvironment().getAuthorizationServletConfig().getHeaderFieldName());
                        info("Got username from header \""+ getServiceEnvironment().getAuthorizationServletConfig().getHeaderFieldName() + "\" + directly: " + x);
                    }

                    if (isEmpty(x)) {
                        if (getServiceEnvironment().getAuthorizationServletConfig().isRequireHeader()) {
                            throw new GeneralException("Error: configuration required using the header \"" +
                                    getServiceEnvironment().getAuthorizationServletConfig().getHeaderFieldName() + "\" " +
                                    "but this was not set. Cannot continue."
                            );
                        }
                        // not required, it is null

                    } else {
                        // name is set. optional or required
                        aState.getTransaction().setUsername(x);
                        info("*** storing user name = " + x);
                        getTransactionStore().save(aState.getTransaction());

                        // make it display pretty as per usual conventions. This is never reused, however.
                        aState.getRequest().setAttribute(AUTHORIZATION_USER_NAME_VALUE, escapeHtml(x));
                    }
                } else {
                    info("*** PRESENT: Use headers DISABLED.");

                }
                JSPUtil.fwd(state.getRequest(), state.getResponse(), initPage);
                info("3.a. User information obtained for grant = " + aState.getTransaction().getAuthorizationGrant());
                break;
            case AUTHORIZATION_ACTION_OK:
                JSPUtil.fwd(state.getRequest(), state.getResponse(), OK_PAGE);
                break;
            default:
        }
    }

    public void handleError(PresentableState state, Throwable t) throws IOException, ServletException {
        AuthorizedState aState = (AuthorizedState) state;
        state.getResponse().setHeader("X-Frame-Options", "DENY");
        state.getRequest().setAttribute("client", aState.getTransaction().getClient());
        JSPUtil.handleException(t, state.getRequest(), state.getResponse(), ERROR_PAGE);
    }

    @Override
    protected void doIt(HttpServletRequest request, HttpServletResponse response) throws Throwable {
        info("*** STARTING request");
        String ag = request.getParameter(TOKEN_KEY);
        DSTransaction trans = null;
        if (ag == null) {
            throw new GeneralException("Error: Invalid request -- no token. Request rejected.");
        }
        trans = getAndCheckTransaction(ag);
        AuthorizedState pState = new AuthorizedState(getState(request), request, response, trans);
        prepare(pState);
        preprocess(new TransactionState(request, response, null, trans));

        switch (pState.getState()) {
            case AUTHORIZATION_ACTION_OK:
                // get the cert and store it. Then forward user.
                try {
                    if (trans.getClient().isProxyLimited()) {
                        doCertRequestLimited(request, response);
                    } else {
                        doCertRequest(request, response);
                    }
                    // There is nothing to present, since the spec requires a redirect
                    // at this point.
                    return;

                } catch (GeneralSecurityException t) {
                    // redo whole attempt
                    info("**** AuthZServlet, got GSX");

                    request.setAttribute(RETRY_MESSAGE, getServiceEnvironment().getMessages().get(RETRY_MESSAGE));
                    pState.setState(AUTHORIZATION_ACTION_START);
                    prepare(pState);
                }
                break;
            case AUTHORIZATION_ACTION_START:
                // no processing needed for initial request.
                break;
            default:
                // nothing to do here either.
        }
        present(pState);
    }

    public int getState(HttpServletRequest request) {
        String action = request.getParameter(AUTHORIZATION_ACTION_KEY);
        log("action = " + action);
        if (action == null || action.length() == 0) return AUTHORIZATION_ACTION_START;
        if (action.equals(AUTHORIZATION_ACTION_OK_VALUE)) return AUTHORIZATION_ACTION_OK;
        throw new DSException("Error: unknown authorization request action = \"" + action + "\"");
    }

    /*
         Get the transaction associated with the authorization grant token and check that it passes sanity
         checks. If so, return it, If not, throw the appropriate exception.
     */
    protected DSTransaction getAndCheckTransaction(String token) throws IOException {
        DateUtils.checkTimestamp(token);
        AuthorizationGrant grant = MyProxyDelegationServlet.getServiceEnvironment().getTokenForge().getAuthorizationGrant(token);
        DSTransaction trans = MyProxyDelegationServlet.getServiceEnvironment().getTransactionStore().get(grant);
        if (trans == null) {
            warn("Error: no delegation request found for " + token);
            throw new DSException("Error: no delegation request found.");
        }
        if (!trans.isAuthGrantValid()) {
            warn("Error: Invalid token " + token);
            throw new DSException("Error: Invalid token");
        }
        checkClient(trans.getClient());
        // ok, we've run the gauntlet and can now try to get the
        return trans;
    }

    /**
     * The actual work for this servlet. Once the information is in hand, go to MyProxy and get the cert.
     *
     * @param request
     * @param response
     * @throws Throwable
     */
    protected void doCertRequest(HttpServletRequest request, HttpServletResponse response) throws Throwable {
        DSTransaction trans = getAndCheckTransaction(request.getParameter(MyProxyDelegationServlet.TOKEN_KEY));
        String userName = trans.getUsername();
        String password = null;
        info("*** stored transaction username = " + userName);
        if (isEmpty(userName)) {
            userName = request.getParameter(AUTHORIZATION_USER_NAME_KEY);
            info("got user name from form = " + userName);

            if (isEmpty(userName)) {
                throw new GeneralException("Error: No user name found");
            }
            password = request.getParameter(AUTHORIZATION_PASSWORD_KEY);
        }

        info("3.b. transaction has user name = " + userName);

        info("3.b. Starting call to MyProxy for username = " + userName);
        // The right place to invoke the pre-processor.
        preprocess(new TransactionState(request, response, null, trans));
        String xx = " transaction =" + trans.getIdentifierString() + " and client=" + trans.getClient().getIdentifierString();
        info("3.b. " + xx);
        Collection<X509Certificate> certs = null;
        Throwable lastException = null;
        info("3.b. MP facade #=" + MyProxyDelegationServlet.getServiceEnvironment().getMyProxyServices().size());
        for (MyProxyServiceFacade msp : MyProxyDelegationServlet.getServiceEnvironment().getMyProxyServices()) {
            info("3.b. myproxy is " + msp.getFacadeConfiguration().getHostname() + ":" + msp.getFacadeConfiguration().getPort());
            try {
                certs = msp.getCerts(
                        userName,
                        password,
                        trans.getCertReq().getDEREncoded(),
                        trans.getLifetime());
            } catch (GeneralSecurityException gsx) {
                error("failed to get cert for token " + xx + ", message = " + gsx.getMessage());
                throw gsx;
            } catch (Throwable t) {
                lastException = t;
                info("3.b. returned exception is " + t.getClass().getName());
            }
        }
        if (certs == null) {
            info("Error: No usable MyProxy service found." + xx);
            throw new GeneralException("Error: No usable MyProxy service found.", lastException);
        }

        info("3.c. Got cert from MyProxy & storing it, " + xx);

        MyX509Certificates myCerts = new MyX509Certificates(certs);
        trans.setProtectedAsset(myCerts);
        // Our response is a simple ok, since otherwise exceptions are thrown. No need to set this since that is the default.
        trans.setVerifier(MyProxyDelegationServlet.getServiceEnvironment().getTokenForge().getVerifier());
        MyProxyDelegationServlet.getServiceEnvironment().getTransactionStore().save(trans);

        // Do callback. // set the verifier.
        debug("4.a. verifier = " + trans.getVerifier() + ", " + xx);
        String cb = trans.getCallback().toString();
        info("4.a. starting redirect to " + cb + ", " + xx);
        // if the callback is overloaded already (which we allow, just append the token and verifier.
        cb = cb + (cb.indexOf("?") == -1 ? "?" : "&") + "oauth_token=" + trans.getIdentifierString() + "&oauth_verifier=" + trans.getVerifier().getToken();
        response.sendRedirect(cb);
        info("4.b. Redirect to callback " + cb + " ok, " + xx);
    }

    protected void doCertRequestLimited(HttpServletRequest request, HttpServletResponse response) throws Throwable {

        DSTransaction trans = getAndCheckTransaction(request.getParameter(MyProxyDelegationServlet.TOKEN_KEY));
        String xx = " transaction =" + trans.getIdentifier() + " and client=" + trans.getClient().getIdentifier();

        String userName = trans.getUsername();
        String password = null;
        info("*** stored transaction username = " + userName);
        if (isEmpty(userName)) {
            userName = request.getParameter(AUTHORIZATION_USER_NAME_KEY);
            info("got user name from form = " + userName);

            if (isEmpty(userName)) {
                throw new GeneralException("Error: No user name found");
            }
            password = request.getParameter(AUTHORIZATION_PASSWORD_KEY);
        }

        info("3.b. transaction has user name = " + userName);

       /* String userName = request.getParameter(AUTHORIZATION_USER_NAME_KEY);
        String password = request.getParameter(AUTHORIZATION_PASSWORD_KEY);*/
        info("3.b. Starting limited cert request call to MyProxy for username = " + userName);
        // DSTransaction trans = getAndCheckTransaction(request.getParameter(MyProxyDelegationServlet.TOKEN_KEY));
        info("3.b. " + xx);

        // Generate a key pair and a cert request to send it to MyProxy server
        KeyPair keyPair = null;
        PKCS10CertificationRequest localCertRequest = null;
        try {
            keyPair = getServiceEnvironment().getKeyPair();
            localCertRequest = CertUtil.createCertRequest(keyPair);
        } catch (GeneralSecurityException e) {
            error("3.b. " + e.getMessage());
        }
        debug("3.b. " + CertUtil.fromCertReqToString(trans.getCertReq()));
        debug("3.b. " + CertUtil.fromCertReqToString(localCertRequest));

        // We assume that MyProxy server returns an EEC signed by a self-signed cert.
        LinkedList<X509Certificate> certs = new LinkedList<X509Certificate>();
        Throwable lastException = null;
        info("3.b. MP facade #=" + MyProxyDelegationServlet.getServiceEnvironment().getMyProxyServices().size());
        for (MyProxyServiceFacade msp : MyProxyDelegationServlet.getServiceEnvironment().getMyProxyServices()) {
            info("3.b. myproxy is " + msp.getFacadeConfiguration().getHostname() + ":" + msp.getFacadeConfiguration().getPort());
            try {
                Collection<X509Certificate> returnedCerts = msp.getCerts(
                        userName,
                        password,
                        localCertRequest.getDEREncoded(),
                        trans.getLifetime());
                certs.addAll(returnedCerts);
            } catch (GeneralSecurityException gsx) {
                error("failed to get cert for token " + xx + ", message = " + gsx.getMessage());
                throw gsx;
            } catch (Throwable t) {
                lastException = t;
                error("3.b. returned exception is " + t.getClass().getName());
            }
        }
        if (certs.isEmpty()) {
            info("Error: No usable MyProxy service found." + xx);
            throw new GeneralException("Error: No usable MyProxy service found.", lastException);
        }

        info("3.c. Got cert from MyProxy, issuing a limited proxy & storing it, " + xx);

        // Sign a cert req from OAuth client using a cert obtained from MyProxy server
        Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
        ProxyPolicy policy = new ProxyPolicy(ProxyPolicy.LIMITED);
        ProxyCertInfo proxyCertInfo = new ProxyCertInfo(policy);
        X509ExtensionSet extSet = new X509ExtensionSet();
        extSet.add(new ProxyCertInfoExtension(proxyCertInfo));

        BouncyCastleCertProcessingFactory factory =
                BouncyCastleCertProcessingFactory.getDefault();
        try {
            // add the cert afterwards so there is no issue with modifying the list early.
            X509Certificate x = factory.createProxyCertificate(certs.getLast(),
                    keyPair.getPrivate(), trans.getCertReq().getPublicKey(),
                    (int) (trans.getLifetime() / 1000),
                    GSI_4_LIMITED_PROXY, extSet, null);
            certs.addFirst(x);
        } catch (GeneralSecurityException e) {
            String errMsg = "3.c. Error: signing a limited proxy credential: " + e.getMessage();
            error(errMsg); // log it
            throw new GeneralException(errMsg, e); // throw it.
        }

        // Storing an EEC and a limited proxy cert
        MyX509Certificates myCerts = new MyX509Certificates(certs);
        trans.setProtectedAsset(myCerts);
        trans.setUsername(userName);
        // Our response is a simple ok, since otherwise exceptions are thrown. No need to set this since that is the default.
        trans.setVerifier(MyProxyDelegationServlet.getServiceEnvironment().getTokenForge().getVerifier());
        MyProxyDelegationServlet.getServiceEnvironment().getTransactionStore().save(trans);

        // Do callback. // set the verifier.
        debug("4.a. verifier = " + trans.getVerifier() + ", " + xx);
        String cb = trans.getCallback().toString();
        info("4.a. starting redirect to " + cb + ", " + xx);
        cb = cb + (cb.indexOf("?") == -1 ? "?" : "&") + "oauth_token=" + trans.getIdentifierString() + "&oauth_verifier=" + trans.getVerifier().getToken();
        response.sendRedirect(cb);
        info("4.b. Redirect to callback " + cb + " ok, " + xx);
    }
}
