package edu.uiuc.ncsa.myproxy.oa4mp.server.servlet;

import edu.uiuc.ncsa.myproxy.oa4mp.server.DSTransaction;
import edu.uiuc.ncsa.security.delegation.server.issuers.ProtectedAssetIssuer;
import edu.uiuc.ncsa.security.delegation.server.request.PARequest;
import edu.uiuc.ncsa.security.delegation.server.request.PAResponse;
import edu.uiuc.ncsa.security.delegation.servlet.TransactionState;
import edu.uiuc.ncsa.security.delegation.token.AccessToken;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.HashMap;

import static edu.uiuc.ncsa.security.core.util.DateUtils.checkTimestamp;

/**
 * <p>Created by Jeff Gaynor<br>
 * on May 18, 2011 at  2:44:47 PM
 */
public class CertServlet extends MyProxyDelegationServlet {
    protected ProtectedAssetIssuer getPAI() throws IOException {
        return getServiceEnvironment().getProtectedAssetIssuer();
    }

    @Override
    protected void doIt(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse) throws Throwable {
        doDelegation(httpServletRequest, httpServletResponse);
    }

    protected void doDelegation(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse) throws Throwable {
        info("6.a. Starting to process cert request");
        PARequest paRequest = new PARequest(httpServletRequest, getClient(httpServletRequest));
        checkClient(paRequest.getClient());
        String cc = "client = " + paRequest.getClient().getIdentifier();
        paRequest.setAccessToken(getServiceEnvironment().getTokenForge().getAccessToken(httpServletRequest));

        PAResponse paResponse = (PAResponse) getPAI().process(paRequest);
        AccessToken accessToken = paResponse.getAccessToken();
        checkTimestamp(accessToken.getToken());
        debug("6.a. " + cc);

        DSTransaction t = null;

        t = (DSTransaction) getTransactionStore().get(accessToken);
        if (t == null) {
            throw new ServletException("Error: no transaction found for access token \"" + accessToken + "\"");
        }

        if (!t.isAccessTokenValid()) {
            throw new ServletException("Error: invalid access token. Request refused");
        }

        info("6.a. Processing request for transaction " +  t.getIdentifier());
        t.setAccessTokenValid(false);
        preprocess(new TransactionState(httpServletRequest, httpServletResponse, paResponse.getParameters(), t));

        debug("6.a. protected asset:" + (t.getProtectedAsset()==null?"(null)":"ok") + ", " + cc);
        HashMap<String, String> username = new HashMap<String, String>();
        username.put("username", t.getUsername());
        username.putAll(paResponse.getParameters());
        paResponse.setAdditionalInformation(username);
        paResponse.setProtectedAsset(t.getProtectedAsset());
        debug("6.a. Added username \"" + t.getUsername() + "\" & cert for request from " + cc);
        getTransactionStore().save(t);

        info("6.b. Done with cert request " + cc);
        paResponse.write(httpServletResponse);
        info("6.b. Completed transaction " + t.getIdentifierString() + ", " + cc);
        postprocess(new TransactionState(httpServletRequest, httpServletResponse, paResponse.getParameters(), t));
    }

}
