package edu.uiuc.ncsa.myproxy.oa4mp.server.servlet;

import edu.uiuc.ncsa.myproxy.oa4mp.server.DSException;
import edu.uiuc.ncsa.myproxy.oa4mp.server.DSTransaction;
import edu.uiuc.ncsa.myproxy.oa4mp.server.ServiceConstantKeys;
import edu.uiuc.ncsa.security.core.exceptions.GeneralException;
import edu.uiuc.ncsa.security.core.exceptions.UnsupportedProtocolException;
import edu.uiuc.ncsa.security.delegation.server.ExceptionWrapper;
import edu.uiuc.ncsa.security.delegation.server.UnapprovedClientException;
import edu.uiuc.ncsa.security.delegation.server.request.AGRequest;
import edu.uiuc.ncsa.security.delegation.server.request.AGResponse;
import edu.uiuc.ncsa.security.delegation.servlet.TransactionState;
import edu.uiuc.ncsa.security.delegation.storage.Client;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.net.URI;
import java.util.Map;

/**
 * <p>Created by Jeff Gaynor<br>
 * on May 17, 2011 at  3:38:58 PM
 */
public class InitServlet extends MyProxyDelegationServlet {
    /**
     * Default is to just call {@link #doDelegation(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)}
     * and return. You may over-ride this to do any pre or post -processing you require.
     *
     * @param httpServletRequest
     * @param httpServletResponse
     * @throws Throwable
     */
    @Override
    protected void doIt(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse) throws Throwable {
        doDelegation(httpServletRequest, httpServletResponse);
    }

    /**
     * Actual work call. This parses and returns the passed in parameters.
     *
     * @param req
     * @param resp
     * @return
     * @throws IOException
     * @throws ServletException
     */
    protected void doDelegation(HttpServletRequest req, HttpServletResponse resp) throws Throwable {
        Client client = getClient(req);
        try {
            checkClient(client);
            String cid = "client=" + client.getIdentifier();
            info("2.a. Starting a new cert request: " + cid);
            AGResponse agResponse = (AGResponse) getAGI().process(new AGRequest(req, client));
            Map<String, String> params = agResponse.getParameters();
            String certReq = params.get(CERT_REQUEST_KEY);
            if (isEmpty(certReq)) {
                throw new DSException("Error: No cert request");
            }
            String certLifetimeString = params.get(CERT_LIFETIME_KEY);
            long certLifetime = 0;
            if (!isEmpty(certLifetimeString)) {
                // set to default
                try {
                    // Again, the lifetime is stored in milliseconds, but the request is in seconds.
                    certLifetime = Long.parseLong(certLifetimeString) * 1000;
                    // Negative value is also rejected. Send along zero ==> take server default.
                    if (certLifetime < 0) {
                        certLifetime = 0L;
                    }
                } catch (NumberFormatException x) {
                    info("No/bad cert lifetime found (" + certLifetimeString + "), using default. " + cid);
                }
            }
            String callback = params.get(getServiceEnvironment().getConstants().get(ServiceConstantKeys.CALLBACK_URI));
            if (isEmpty(callback)) {
                throw new GeneralException("Error: No callback specified");
            }
            URI cb = URI.create(callback); // if it is legal, continue, otherwise, this throws a runtime exception

            if (cb.getScheme() == null || !cb.getScheme().equals("https")) {
                throw new UnsupportedProtocolException("Error: protocol in the callback must be https");
            }
            // Now we have to add back in any unrecognized parameters, as per the spec.

            DSTransaction transaction = newTransaction();
            transaction.setAuthorizationGrant(agResponse.getGrant());
            debug("creating transaction for " + cid + ", trans id=" + transaction.getIdentifierString());

            transaction.setClient(client);
            transaction.setAuthGrantValid(true);
            transaction.setAccessTokenValid(false);
            transaction.setCallback(URI.create(callback));
            transaction.setCertReq(certReq);
            transaction.setLifetime(certLifetime);
            getTransactionStore().save(transaction);
            preprocess(new TransactionState(req, resp, params, transaction));
            debug("saved transaction for " + cid + ", trans id=" + transaction.getIdentifierString());
            params.remove(CERT_REQUEST_KEY);
            params.remove(CERT_LIFETIME_KEY);
            params.remove(CONSUMER_KEY);

            agResponse.write(resp);
            info("2.b finished initial request for token =\"" + transaction.getIdentifierString() + "\".");

            postprocess(new TransactionState(req, resp, params, transaction));
        } catch (Throwable t) {
            if (t instanceof UnapprovedClientException) {
                throw t;
            }
            ExceptionWrapper ew = new ExceptionWrapper(t, client);
            // stick on the client and throw it.
            throw ew;
        }
    }
}
