package edu.uiuc.ncsa.myproxy.oa4mp.server.servlet;

import edu.uiuc.ncsa.myproxy.oa4mp.server.DSTransaction;
import edu.uiuc.ncsa.myproxy.oa4mp.server.ServiceEnvironment;
import edu.uiuc.ncsa.myproxy.oa4mp.server.ServiceEnvironmentImpl;
import edu.uiuc.ncsa.myproxy.oa4mp.server.util.*;
import edu.uiuc.ncsa.security.core.Identifier;
import edu.uiuc.ncsa.security.core.cache.Cleanup;
import edu.uiuc.ncsa.security.core.cache.ValidTimestampPolicy;
import edu.uiuc.ncsa.security.core.exceptions.GeneralException;
import edu.uiuc.ncsa.security.core.exceptions.UnknownClientException;
import edu.uiuc.ncsa.security.core.util.BasicIdentifier;
import edu.uiuc.ncsa.security.delegation.server.ExceptionWrapper;
import edu.uiuc.ncsa.security.delegation.server.UnapprovedClientException;
import edu.uiuc.ncsa.security.delegation.server.issuers.AccessTokenIssuer;
import edu.uiuc.ncsa.security.delegation.server.issuers.AuthorizationGrantIssuer;
import edu.uiuc.ncsa.security.delegation.servlet.TransactionFilter;
import edu.uiuc.ncsa.security.delegation.servlet.TransactionState;
import edu.uiuc.ncsa.security.delegation.storage.Client;
import edu.uiuc.ncsa.security.delegation.storage.TransactionStore;
import edu.uiuc.ncsa.security.delegation.storage.impl.BasicTransaction;
import edu.uiuc.ncsa.security.delegation.token.AuthorizationGrant;
import edu.uiuc.ncsa.security.servlet.AbstractServlet;
import edu.uiuc.ncsa.security.servlet.JSPUtil;
import edu.uiuc.ncsa.security.servlet.NotificationListener;
import edu.uiuc.ncsa.security.servlet.mail.MailUtil;
import edu.uiuc.ncsa.security.util.pkcs.KeyPairPopulationThread;
import org.apache.commons.codec.binary.Base64;

import javax.servlet.ServletException;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * <p>Created by Jeff Gaynor<br>
 * on May 17, 2011 at  3:46:53 PM
 */
public abstract class MyProxyDelegationServlet extends AbstractServlet implements TransactionFilter {
    // Same as in client.
    public static final String CLIENT_DEBUG_KEY = "oauth_debug";

    static List<NotificationListener> notificationListeners = new ArrayList<NotificationListener>();

    public static void addNotificationListener(NotificationListener notificationListener) {
        if (!notificationListeners.contains(notificationListener)) {
            notificationListeners.add(notificationListener);
        }
    }

    public static boolean removeNotificationListener(NotificationListener notificationListener) {
        return notificationListeners.remove(notificationListener);
    }

    // individual servlets fire events and filter from the list.

    static Cleanup<String, BasicTransaction> transactionCleanup;

    static KeyPairPopulationThread kpt;

    public static final String CERT_REQUEST_KEY = "certreq";
    public static final String CERT_LIFETIME_KEY = "certlifetime";

    /**
     * The key for the token.
     */
    public static final String TOKEN_KEY = "oauth_token";
    public static final String CONSUMER_KEY = "oauth_consumer_key";


    public ServiceEnvironmentImpl loadProperties2() throws IOException {
        ServiceEnvironmentImpl se2 = (ServiceEnvironmentImpl) getConfigurationLoader().load();
        if (se2.isPollingEnabled()) {
            caThread = se2.getClientApprovalThread();
        }
        kpt = new KeyPairPopulationThread(se2.getKeyPairQueue());
        return se2;
    }


    @Override
    public void loadEnvironment() throws IOException {
        if (environment == null) {
            setEnvironment(loadProperties2());
        }
    }


    protected AuthorizationGrantIssuer getAGI() throws IOException {
        return getServiceEnvironment().getAuthorizationGrantIssuer();
    }

    protected AccessTokenIssuer getATI() throws IOException {
        return getServiceEnvironment().getAccessTokenIssuer();
    }


    public static ServiceEnvironment getServiceEnvironment() {
        return (ServiceEnvironment) getEnvironment();
    }

    static boolean notifiersSet = false;

    String getTemplate(File filename) throws IOException {

        String body = "";
        try {
            FileInputStream fr = new FileInputStream(filename);
            StringBuffer sb = new StringBuffer();
            int z = 0;
            while ((z = fr.read()) != -1) {
                sb.append((char) z);
            }
            body = sb.toString();
            fr.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return body;
    }

    protected void setupNotifiers() throws IOException {
        // do this once or you will have a message sent for each listener!
        if (notifiersSet) return;

        // debugging this...
        NewClientNotifier newClientNotifier = new NewClientNotifier(getServiceEnvironment().getMailUtil(), getMyLogger());
        addNotificationListener(newClientNotifier);
        MailUtil x = new MailUtil(getServiceEnvironment().getMailUtil().getMailEnvironment());
        String fName = getServletContext().getInitParameter("oa4mp:server.error.subject");
        if (fName == null) {
            info("No error notification subject set. Skipping...");
            notifiersSet = true;
            return;
        }
        x.setSubjectTemplate(getTemplate(new File(fName)));

        fName = getServletContext().getInitParameter("oa4mp:server.error.message");
        if (fName == null) {
            info("No error notification message body set. Skipping...");
            notifiersSet = true;
            return;
        }

        x.setMessageTemplate(getTemplate(new File(fName)));

        ExceptionEventNotifier exceptionNotifier = new ExceptionEventNotifier(x, getMyLogger());
        addNotificationListener(exceptionNotifier);
        notifiersSet = true;
    }

    @Override
    public void init() throws ServletException {
        super.init();

        if (transactionCleanup == null) {
            transactionCleanup = new Cleanup<String, BasicTransaction>(getMyLogger());
            transactionCleanup.setStopThread(false);
            try {
                transactionCleanup.setMap(getTransactionStore());
            } catch (IOException e) {
                throw new GeneralException("Error: getting transaction store");
            }
            transactionCleanup.addRetentionPolicy(new ValidTimestampPolicy());
            transactionCleanup.start();
            info("Starting transaction store cleanup thread");
        }
        if (caThread != null && !caThread.isAlive()) {
            caThread.setStopThread(false);
            caThread.start();
        }
        if (kpt != null && !kpt.isAlive()) {
            kpt.setStopThread(false);
            kpt.start();
        }
        try {
            setupNotifiers();
        } catch (IOException e) {
            throw new GeneralException("Error: could not set up notifiers ", e);
        }
    }

    static AbstractCLIApprover.ClientApprovalThread caThread = null;

    protected void shutdownCleanup(Cleanup c) {
        if (c != null && !c.isStopThread()) {
            c.setStopThread(true); // Just in case...
            c.interrupt();
        }
    }

    @Override
    public void destroy() {
        super.destroy();
        shutdownCleanup(transactionCleanup);
        if (caThread != null) {
            caThread.setStopThread(true);
        }
        if (kpt != null) {
            kpt.setStopThread(true);
        }
    }

    protected TransactionStore getTransactionStore() throws IOException {
        return getServiceEnvironment().getTransactionStore();
    }

    public Client getClient(HttpServletRequest req) throws ServletException, IOException {
        Identifier id = BasicIdentifier.newID(req.getParameter(CONSUMER_KEY));
        return getClient(id);
    }

    public Client getClient(Identifier identifier) throws ServletException, IOException {
        if (identifier == null) {
            throw new UnknownClientException("no client id");
        }
        Client c = getServiceEnvironment().getClientStore().get(identifier);
        if (c == null) {
            throw new UnknownClientException("client not found", identifier);
        }

        return c;
    }

    public DSTransaction newTransaction() throws IOException {
        return getServiceEnvironment().getTransactionStore().create();
    }

    protected DSTransaction getTransaction(AuthorizationGrant grant) throws IOException {
        return (DSTransaction) getTransactionStore().get(grant);
    }

    /**
     * A utility to get the client from the authorization grant. This looks up the transaction
     *
     * @param authorizationGrant
     * @return
     */
    protected Client getClient(AuthorizationGrant authorizationGrant) throws IOException {
        DSTransaction transaction = getTransaction(authorizationGrant);
        return transaction.getClient();
    }

    /**
     * Checks if the client is approved. This should be done before each leg of the process
     *
     * @param client
     */
    public void checkClient(Client client) throws IOException {
        if (!getServiceEnvironment().getClientApprovalStore().isApproved(client.getIdentifier())) {
            String ww = "The client with identifier \"" + client.getIdentifier() + "\" has not been approved. Request rejected. Please contact your administrator.";
            warn(ww);
            throw new UnapprovedClientException("Error: " + ww, client);
        }
    }

    protected boolean isEmpty(String x) {
        return x == null || x.length() == 0;
    }


    /**
     * Note that if you override this, you should call super, which sets some security-related headers, but touches nothing else.
     *
     * @param state
     * @throws Throwable
     */
    @Override
    public void preprocess(TransactionState state) throws Throwable {
        state.getResponse().setHeader("X-Frame-Options", "DENY");
    }

    @Override
    public void postprocess(TransactionState state) throws Throwable {
    }

    /**
     * The server generates a redirect suitable for consumption by the ErrorServlet.
     * The client will also parse the redirect in case there is no error handler on the server.
     * This is currently a beta feature only.
     *
     * @param t
     * @param request
     * @param response
     * @throws IOException
     * @throws ServletException
     */
    protected void setRedirect(Throwable t, HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        if (isDebugOn()) {
            t.printStackTrace();
        }
        String x = getServiceEnvironment().getAuthorizationGrantIssuer().getAddress().toString();
        if (!x.endsWith("/")) {
            x = x + "/";
        }
        response.setHeader("X-Frame-Options", "DENY");
        Throwable cause = t;
        Client client = null;
        String identifier = "";

        if (t instanceof ExceptionWrapper) {
            ExceptionWrapper ew = (ExceptionWrapper) t;
            cause = ew.getCause();
            client = ew.getClient();
            if (client.getIdentifier() != null) {
                identifier = client.getIdentifierString();
            }
        }
        String causeName = cause.getClass().getSimpleName();
        String message = URLEncoder.encode(cause.getMessage(), "UTF-8");
        // unknown or unapproved client exceptions are thrown before any other exceptions. These
        // contain the identifier which is available no other way, so these are handled as special cases.
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PrintStream printStream = new PrintStream(baos);
        cause.printStackTrace(printStream);
        printStream.flush();
        printStream.close();
        String stackTrace = Base64.encodeBase64URLSafeString(baos.toByteArray());
        if (cause instanceof UnknownClientException) {
            UnknownClientException uce = (UnknownClientException) t;
            identifier = uce.getIdentifier().toString();
            response.sendRedirect(x + "error"
                    + "?" + ErrorServlet.CAUSE + "=" + causeName
                    + "&" + ErrorServlet.STACK_TRACE + "=" + stackTrace
                    + "&" + ErrorServlet.IDENTIFIER + "=" + identifier
                    + "&" + ErrorServlet.MESSAGE + "=" + message);
            return;
        }
        if (cause instanceof UnapprovedClientException) {
            UnapprovedClientException uce = (UnapprovedClientException) t;
            identifier = uce.getClient().getIdentifier().toString();
            response.sendRedirect(x + "error"
                    + "?" + ErrorServlet.CAUSE + "=" + causeName
                    + "&" + ErrorServlet.STACK_TRACE + "=" + stackTrace
                    + "&" + ErrorServlet.IDENTIFIER + "=" + identifier
                    + "&" + ErrorServlet.MESSAGE + "=" + message);
            return;
        }


        response.sendRedirect(x + "error"
                + "?" + ErrorServlet.CAUSE + "=" + causeName
                + "&" + ErrorServlet.IDENTIFIER + "=" + identifier
                + "&" + ErrorServlet.STACK_TRACE + "=" + stackTrace
                + "&" + ErrorServlet.MESSAGE + "=" + message);

    }

    protected void fireExceptionCaught(Throwable throwable, Map<String, String> state) {
        for (NotificationListener notificationListener : notificationListeners) {
            if (notificationListener instanceof ExceptionEventListener) {
                ((ExceptionEventListener) notificationListener).fireExceptionCaught(new ExceptionEvent(this, throwable, state));
            }
        }
    }

    /**
     * Have this display an error page and send a message to the admin.
     *
     * @param t
     * @param request
     * @param response
     * @throws IOException
     * @throws ServletException
     */
    @Override
    protected void handleException(Throwable t, HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        if (getServiceEnvironment().isDebugOn()) {
            t.printStackTrace();
        }
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PrintWriter pw = new PrintWriter(baos);
        t.printStackTrace(pw);
        pw.flush();
        pw.close();
        String stacktrace = new String(baos.toByteArray());
        error("INTERNAL ERROR: " + (t.getMessage() == null ? "(no message)" : t.getMessage())); // log it appropriately
        error(stacktrace); // print out the stack trace in the log.

        String clientDebug = request.getParameter(CLIENT_DEBUG_KEY);
        boolean doClientDebug = false;
        doClientDebug = clientDebug != null && Boolean.parseBoolean(clientDebug);
        request.setAttribute("message", t.getMessage() + "\n");

        if (doClientDebug) {
            // ONLY in cases where a client *specifically* requests it return this sort of information.
            // The intent is to allow low-level **server** debugging for really thorny issues.
            HashMap<String, String> state = new HashMap<String, String>();
            state.put("servletEngine", getServletContext().getServerInfo() + " ");
            state.put("host", request.getServerName() + ":" + request.getServerPort() + "\n");
            state.put("client", request.getRemoteHost() + "\n");
            state.put("servlet", getServletName() + "\n");
            state.put("url", request.getRequestURI() + "\n");
            Cookie[] cookies = request.getCookies();
            String cc = "";
            if (cookies == null) {
                cc = "(none)";
            } else {
                for (Cookie cookie : cookies) {
                    cc = cc + cookie.getName() + " : " + cookie.getValue() + "\n";
                }
            }
            state.put("cookies", cc);


            state.put("message", t.getMessage() + "\n");
            state.put("stackTrace", stacktrace);
            fireExceptionCaught(t, state);
            setRedirect(t, request, response);
        }else{
            JSPUtil.fwd(request, response, "/oops.jsp");
        }
    }
}
